#include "arith.h"
#include "error.h"
#include "defs.h"

#include <stdio.h>
#include <assert.h>
#include <string.h>

#define ARITH_BUFF_SIZE    (1 << 18)
#define MIN(a, b) ((a) < (b) ? (a) : (b))

/* variabile statice :) */
static uint32_t low, high;
static uint32_t code;                         /* codul curent */
static uint32_t underflow_bits;
static uint32_t encode_wrote;
static uint32_t decode_read;

/* buffer */
static uint8_t buffer[ARITH_BUFF_SIZE];       /* bufferul propriu-zis */
static uint8_t byte;
static uint32_t where;                        /* byte-ul si pozitia bitulyu curent */
static uint32_t filled;                       /* cat este plin din buffer */

/* functia pentru output, input */
static void (*flush_out)(const uint8_t *, uint32_t);       /* bufferul si marimea sa */
static void (*flush_in)(uint8_t *, uint32_t *);            /* bufferul si marimea sa maxima */

/* functii locale, encode */

static void encode_flush(void)
{
    assert(flush_out != NULL);

    flush_out(buffer, filled);
    encode_wrote += filled;
    filled = 0;
}

static void encode_output_bit(uint32_t bit)
{
    if(bit != 0)
        byte |= 1 << where;

    if(++ where == (uint32_t)8) {        /* am umplut un byte, il punem in buffer */
        if(filled == ARITH_BUFF_SIZE)    /* e plin bufferul? */
            encode_flush();

        buffer[filled ++] = byte;
        where = 0; byte = 0;
    }
}

/* functii globale, encode */

uint32_t arith_encode_wrote(void)
{
    return encode_wrote;
}

void arith_encode_init(void (*f)(const uint8_t *, uint32_t))
{
    low = (uint32_t)0x00000000;
    high = (uint32_t)0xffffffff;
    underflow_bits = 0;

    flush_out = f;
    encode_wrote = 0;

    filled = 0;
    where = 0;
    byte = 0;
}

void arith_encode_finish(void)
{
    encode_output_bit(low & 0x40000000);
    underflow_bits ++;

    while(underflow_bits -- > 0)
        encode_output_bit(~low & 0x40000000);

    /* trebuie sa umplem byte-ul */
    while(where != 0)
        encode_output_bit(0);

    encode_flush();
}

void arith_encode(uint32_t low_count, uint32_t high_count,
                  uint32_t scale)
{
    uint64_t range;

    /* rescale high and low for the new symbol */
    range = (uint64_t)(high - low) + 1;
    high = low + (uint32_t)(range*high_count / scale - 1);
    low = low + (uint32_t)(range*low_count / scale);

    /*
    fprintf(stderr, "lc = %X, hc = %X  ", low_count, high_count);
    fprintf(stderr, "lo = %X, hi = %X\n", low, high);
    */

    /* high and low should be stabilezed */
    while((low & 0x80000000) == (high & 0x80000000)) {
        encode_output_bit(high & 0x80000000);

        while(underflow_bits > 0) {
            encode_output_bit(~high & 0x80000000);
            underflow_bits --;
        }

        high <<= 1; high |= 1;
        low <<= 1;
    }

    /* check for underflow */
    while((low & 0x40000000) != 0 && (high & 0x40000000) == 0) {
        underflow_bits ++;

        /* eliminam bitul al 2-lea cel mai semnificativ */
        /* avem: low -> 01..., hi -> 10... */
        low <<= 1; low &= 0x7fffffff;
        high <<= 1; high |= 0x80000001;
    }
}

/* functii locale, decode */

static void decode_flush(void)
{
    size_t i;

    /* daca nu mai avem ce citi, dam garbage */
    if(encode_wrote == decode_read) {
        filled = ARITH_BUFF_SIZE;
        memset(buffer, 0x3f, filled);
        return;
    }

    filled = MIN(encode_wrote - decode_read, ARITH_BUFF_SIZE);
    flush_in(buffer, &filled);
    decode_read += filled;

    assert(filled != 0);
    error_print(3, "decode flushed %d byte(s)", filled);

    /* intorc buffer-ul ca sa nu mai folosesc O variabila */
    for(i = 0; i < filled / 2; i ++) {
        uint8_t temp = buffer[filled-1-i];
        buffer[filled-1-i] = buffer[i];
        buffer[i] = temp;
    }
}

static int decode_input_bit(void)
{
    int value;

    if(where == (uint32_t)256) {           /* urmatorul byte? */
        if(filled == 0) decode_flush();    /* trebuie sa citim */
        byte = buffer[-- filled];
        where = (uint32_t)1;
    }

    value = (int)(byte & where);
    where <<= 1;

    if(value != 0) return 1;
    return 0;
}

/* functii globale, decode */

uint32_t arith_decode_read(void)
{
    return decode_read;
}

void arith_decode_init(void (*f)(uint8_t *, size_t *), uint32_t wrote)
{
    int i;

    low = 0x00000000;
    high = (uint32_t)0xffffffff;

    flush_in = f;

    /* ne obligam sa citim */
    filled = 0;
    where = (uint32_t)256;
    decode_read = 0;
    encode_wrote = wrote;

    /* initializam code */
    code = 0;
    for(i = 0; i < 32; ++ i)
        code = (code << 1) + decode_input_bit();

    /*
    fprintf(stderr, "init: code = %d\n", code);
    */
}

void arith_decode_get_code(uint32_t *count, uint32_t scale)
{
    uint64_t range;
    uint64_t temp;

    range = (uint64_t)(high - low) + 1;
    temp = ((uint64_t)(code - low) + 1)*scale - 1;
    *count = (uint32_t)(temp / range);
}

void arith_decode_remove(uint32_t low_count, uint32_t high_count,
                         uint32_t scale)
{
    uint64_t range;

    /* rescale high and low for the symbol removal */
    range = (uint64_t)(high - low) + 1;
    high = low + (uint32_t)(range*high_count / scale - 1);
    low = low + (uint32_t)(range*low_count / scale);

    /*
    fprintf(stderr, "lc = %X, hc = %X  ", low_count, high_count);
    fprintf(stderr, "lo = %X, hi = %X\n", low, high);
    */

    /* any possible bits are shipped out */
    while((low & 0x80000000) == (high & 0x80000000)) {
        high <<= 1; high |= 1;
        low <<= 1; code <<= 1;
        code += decode_input_bit();
    }

    /* underflow? */
    while((low & 0x40000000) != 0 && (high & 0x40000000) == 0) {
        /* eliminam bitul al 2-lea cel mai semnificativ */
        /* avem: low -> 01..., high -> 10... */
        low <<= 1; low &= 0x7fffffff;
        high <<= 1; high |= 0x80000001;

        code ^= 0x40000000;
        code <<= 1;
        code += decode_input_bit();
    }
}
